/* COPYING ******************************************************************
For copyright and licensing terms, see the file named COPYING.
// **************************************************************************
*/

#if !defined(INCLUDE_REDO_H)
#define INCLUDE_REDO_H

#include <vector>
#include <list>
#include <string>
#include <iosfwd>
#include <cstdio>

extern bool debug;

extern
std::ostream &
msg (
	const char * prog,
	const char * prefix
) ;

#endif
