/* COPYING ******************************************************************
For copyright and licensing terms, see the file named COPYING.
// **************************************************************************
*/

#if !defined(INCLUDE_JOBSERVER_H)
#define INCLUDE_JOBSERVER_H

#include <vector>
#include <list>
#include <string>
#include <cstdio>

/* The GNU make jobserver ***************************************************
// **************************************************************************
*/

namespace jobserver {

extern
bool
try_procure_slot(
	const char * prog
) ;

extern
bool
procure_slot(
	const char * prog
) ;

extern
void
vacate_slot(
	const char * prog
) ;

extern
int
create_pipe(
);

extern
int
fds[2];

}

#endif
